import { Observable, Unsubscribable } from 'rxjs';
import { Event } from './events';
import { Message, NotificationMessage, ResponseError } from './messages';
import { Tracer } from './trace';
import { MessageReader, MessageWriter } from './transport';
export interface Logger {
    error(message: string): void;
    warn(message: string): void;
    info(message: string): void;
    log(message: string): void;
}
export declare enum ConnectionErrors {
    /**
     * The connection is closed.
     */
    Closed = 1,
    /**
     * The connection got unsubscribed (i.e., disposed).
     */
    Unsubscribed = 2,
    /**
     * The connection is already in listening mode.
     */
    AlreadyListening = 3
}
export declare class ConnectionError extends Error {
    readonly code: ConnectionErrors;
    constructor(code: ConnectionErrors, message: string);
}
type HandlerResult<R, E> = R | ResponseError<E> | Promise<R> | Promise<ResponseError<E>> | Promise<R | ResponseError<E>> | Observable<R>;
type StarRequestHandler = (method: string, params?: any, signal?: AbortSignal) => HandlerResult<any, any>;
type GenericRequestHandler<R, E> = (params?: any, signal?: AbortSignal, methodName?: string) => HandlerResult<R, E>;
type StarNotificationHandler = (method: string, params?: any) => void;
type GenericNotificationHandler = (params: any) => void;
export interface Connection extends Unsubscribable {
    sendRequest<R>(method: string, params?: any[], signal?: AbortSignal, toSecondary?: boolean): Promise<R>;
    observeRequest<R>(method: string, params?: any[]): Observable<R>;
    onRequest<R, E>(method: string, handler: GenericRequestHandler<R, E>): void;
    onRequest(handler: StarRequestHandler): void;
    sendNotification(method: string, params?: any[], toSecondary?: boolean): void;
    onNotification(method: string, handler: GenericNotificationHandler): void;
    onNotification(handler: StarNotificationHandler): void;
    trace(tracer: Tracer | null): void;
    onError: Event<[Error, Message | undefined, number | undefined]>;
    onClose: Event<void>;
    onUnhandledNotification: Event<NotificationMessage>;
    listen(): void;
    onUnsubscribe: Event<void>;
}
export interface MessageTransports {
    reader: MessageReader;
    writer: MessageWriter;
}
export declare function createConnection(transports: MessageTransports, logger?: Logger): Connection;
export {};
