import {AbstractMessageReader,AbstractMessageWriter}from'./transport.js';/**
 * Creates a pair of message transports that are connected to each other. One can be used as the server and the
 * other as the client.
 */
function createMessageTransports() {
    const { reader: reader1, writer: writer1 } = createMessagePipe();
    const { reader: reader2, writer: writer2 } = createMessagePipe();
    return [{ reader: reader1, writer: writer2 }, { reader: reader2, writer: writer1 }];
}
/** Creates a single set of transports that are connected to each other. */
function createMessagePipe() {
    let readerCallback;
    const reader = new class extends AbstractMessageReader {
        listen(callback) {
            readerCallback = callback;
        }
    }();
    const writer = new class extends AbstractMessageWriter {
        write(msg) {
            if (readerCallback) {
                readerCallback(msg);
            }
            else {
                throw new Error('reader has no listener');
            }
        }
    }();
    return { reader, writer };
}export{createMessagePipe,createMessageTransports};